@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-cyrillic-ext-400-normal.dd55ea0a.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+0460-052f,U+1c80-1c88,U+20b4,U+2de0-2dff,U+a640-a69f,U+fe2e-fe2f}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-cyrillic-400-normal.3605d18d.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+0400-045f,U+0490-0491,U+04b0-04b1,U+2116}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-greek-ext-400-normal.bc7ace6e.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+1f??}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-greek-400-normal.352cc77a.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+0370-03ff}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-vietnamese-400-normal.52cebac0.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+0102-0103,U+0110-0111,U+0128-0129,U+0168-0169,U+01a0-01a1,U+01af-01b0,U+1ea0-1ef9,U+20ab}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-latin-ext-400-normal.dcc07bcf.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+0100-024f,U+0259,U+1e??,U+2020,U+20a0-20ab,U+20ad-20cf,U+2113,U+2c60-2c7f,U+a720-a7ff}@font-face{font-family:"Roboto";font-style:normal;font-display:swap;font-weight:400;src:url(/static/media/roboto-latin-400-normal.176f8f5b.woff2) format("woff2"),url(/static/media/roboto-all-400-normal.b5a0a0df.woff) format("woff");unicode-range:U+00??,U+0131,U+0152-0153,U+02bb-02bc,U+02c6,U+02da,U+02dc,U+2000-206f,U+2074,U+20ac,U+2122,U+2191,U+2193,U+2212,U+2215,U+feff,U+fffd}.uplot,.uplot *,.uplot :after,.uplot :before{box-sizing:border-box}.uplot{font-family:system-ui,-apple-system,"Segoe UI",Roboto,"Helvetica Neue",Arial,"Noto Sans",sans-serif,"Apple Color Emoji","Segoe UI Emoji","Segoe UI Symbol","Noto Color Emoji";line-height:1.5;width:-webkit-min-content;width:min-content}.u-title{text-align:center;font-size:18px;font-weight:700}.u-wrap{position:relative;-webkit-user-select:none;-ms-user-select:none;user-select:none}.u-over,.u-under{position:absolute}.u-under{overflow:hidden}.uplot canvas{display:block;position:relative;width:100%;height:100%}.u-legend{font-size:14px;margin:auto;text-align:center}.u-inline{display:block}.u-inline *{display:inline-block}.u-inline tr{margin-right:16px}.u-legend th{font-weight:600}.u-legend th>*{vertical-align:middle;display:inline-block}.u-legend .u-marker{width:1em;height:1em;margin-right:4px;background-clip:content-box!important}.u-inline.u-live th:after{content:":";vertical-align:middle}.u-inline:not(.u-live) .u-value{display:none}.u-series>*{padding:4px}.u-series th{cursor:pointer}.u-legend .u-off>*{opacity:.3}.u-select{background:rgba(0,0,0,.07)}.u-cursor-x,.u-cursor-y,.u-select{position:absolute;pointer-events:none}.u-cursor-x,.u-cursor-y{left:0;top:0;will-change:transform;z-index:100}.u-hz .u-cursor-x,.u-vt .u-cursor-y{height:100%;border-right:1px dashed #607d8b}.u-hz .u-cursor-y,.u-vt .u-cursor-x{width:100%;border-bottom:1px dashed #607d8b}.u-cursor-pt{position:absolute;top:0;left:0;border-radius:50%;pointer-events:none;will-change:transform;z-index:100;background-clip:content-box!important}.u-cursor-pt.u-off,.u-cursor-x.u-off,.u-cursor-y.u-off,.u-select.u-off{display:none}
/*# sourceMappingURL=2.c2136c50.chunk.css.map */